import pickle
import os

class OutputLoader():
    def save(obj, path):
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as handle:
            pickle.dump(obj, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def load(path):
        with open(path, 'rb') as handle:
            obj = pickle.load(handle)
        return obj



import numpy as np
import networkx as nx
import torch
import torch.nn as nn
import copy

from nsp.transformer import Transformer

class NNGraph(nx.DiGraph):
    # INCL set to True: includes edge even if weight = 0.
    # Otherwise only edges with weight !=0 are added.
    # Especially, INCL = False will only have one incoming edge per MaxPool Activation
    def __init__(self, activations, INCL=True):
        super().__init__()
        self.INCL = INCL
        self.add_nodes_from_activations(activations)
        self.add_edges_from_jacobian(activations)
        self.inv_F = None
        self.lu_piv = None

    def add_nodes_from_activations(self, activations):
        for index_layer, layeractivation in enumerate(activations.layeractivations):
            for index_generator in np.ndindex(layeractivation.shape):
                self.add_node((index_layer, *index_generator))

    def add_edges_from_jacobian(self, activations):
    # only accept nn.Linear layers if input is flattened
        for index_previous_layer, (layeractivation, layerdescription) in enumerate(zip(activations.layeractivations[1:], activations.layerdescriptions[1:])):
            previous_layeractivation = activations.layeractivations[index_previous_layer]

            if (isinstance(layerdescription, nn.Linear)):
                # assume only flattened inputs x are passed to Linear layers (x = x.view(x.size(0), -1))
                for index_generator_from_node in np.ndindex(previous_layeractivation.shape):
                        for index_generator_to_node in np.ndindex(layeractivation.shape):
                            from_node = (index_previous_layer, *index_generator_from_node)
                            to_node = (index_previous_layer + 1, *index_generator_to_node)
                            assert self.has_node(from_node), f'from_node does not exist: {from_node}'
                            assert self.has_node(to_node), f'to_node does not exist: {to_node}'
                            self.add_edge(from_node, to_node)

            else:
                if (self.INCL): # INCLUDE zero-weight edges
                    if (isinstance(layerdescription, nn.MaxPool2d)):
                        assert (layerdescription.ceil_mode == False)
                        n_channels = previous_layeractivation.shape[0]
                        layerdescription_for_jacobian = nn.Conv2d(in_channels=n_channels, out_channels=n_channels,
                                                                    kernel_size=layerdescription.kernel_size,
                                                                    stride=layerdescription.stride,
                                                                    padding=layerdescription.padding,
                                                                    dilation=layerdescription.dilation,
                                                                    groups=n_channels) # important
                    else:
                      layerdescription_for_jacobian = copy.deepcopy(layerdescription)
                    # draw edge in covergraph even if weight = 0
                    # i.e. ensure jacobian is either 1 (is input) or 0 (is not input) regardless of weight of this input
                    if not (isinstance(layerdescription, nn.AvgPool2d)): # change to if instance has weight attribute then fill with 1
                        layerdescription_for_jacobian.weight.data.fill_(1)
                    # previous_layeractivation.unsqueeze(0) adds the batch dimension
                    jacobian = torch.autograd.functional.jacobian(layerdescription_for_jacobian, previous_layeractivation.clone().detach().unsqueeze(0))
                else: # EXCLUDE zero-weight edges
                    jacobian = torch.autograd.functional.jacobian(layerdescription, previous_layeractivation)
                for index_generator in np.ndindex(jacobian.size()):
                    to_batch, to_channel, to_i, to_j, from_batch, from_channel, from_i, from_j = index_generator
                    if jacobian[index_generator]:
                        from_node = (index_previous_layer, from_channel, from_i, from_j)
                        to_node = (index_previous_layer + 1, to_channel, to_i, to_j)
                        assert self.has_node(from_node), f'from_node does not exist: {from_node}'
                        assert self.has_node(to_node), f'to_node does not exist: {to_node}'
                        self.add_edge(from_node, to_node)

    def get_transformer(self, type='standard'):
        if type =='standard':
            self.inv_F = Transformer.get_inv_fourier_matrix(self, type=type)
            return self.inv_F
        elif type =='laplacian':
            self.lu_piv = Transformer.get_inv_fourier_matrix(self, type=type)
            return self.lu_piv
        else:
            raise NotImplementedError

    def transform(self, activations, type='standard'):
        if type =='standard':
            if self.inv_F is None:
                self.get_transformer(type=type)
            return Transformer.get_fourier_coefficients(activations=activations, inv_F=self.inv_F, type=type, lu_piv=None)
        elif type =='laplacian':
            if self.lu_piv is None:
                self.lu_piv = Transformer.get_inv_fourier_matrix(self, type=type)
            return Transformer.get_fourier_coefficients(activations=activations, inv_F=None, type=type, lu_piv=lu_piv)
        else:
            raise NotImplementedError

    def node_id(self, node):
        nodes = list(self.nodes)
        return nodes.index(node)



import networkx as nx
import numpy as np
import scipy.sparse

class Transformer:
    def get_laplacian(graph):
        undir_graph = graph.to_undirected()
        degrees = [val for (node, val) in sorted(undir_graph.degree(), key=lambda pair: pair[0])]
        L = - nx.adjacency_matrix(undir_graph) # L = -W
        L.setdiag(degrees) # L = D-, no loops
        return L

    def get_inv_fourier_matrix(graph, type='standard'):
        if  type == 'standard':
            nodes = list(graph.nodes)
            n = graph.number_of_nodes()
            inv_F = np.identity(n)
            for node in nodes:
                for incoming_node, _ in graph.in_edges(node):
                    inv_F[nodes.index(node)] = np.logical_or(inv_F[nodes.index(node)], inv_F[nodes.index(incoming_node)])
            inv_F = scipy.sparse.tril(inv_F, format='csr')
        elif type == 'laplacian':
            # compute non-normalized graph laplacian matrix (Paper: The Emerging Field of Signal Processing on Graphs Paper, p.3/B L := D-W)
            L = Transformer.get_laplacian(graph)
            n = graph.number_of_nodes()
            eigenvalues, eigenvectors = np.linalg.eigh(L.asfptype().toarray())
            inv_F = scipy.linalg.lu_factor(eigenvectors)
        else:
            raise NotImplementedError
        return inv_F

    def get_moebius_sum(F, inv_F, x1, x2):
        sum = 0
        for y in range(x2, x1): # x2, x2+1, ... , x1-2, x1-1 ==> y < x1
            if inv_F[v1][v2] and inv_F[v2][v1]:
                sum = sum + F[y][x2]
        return sum

    def get_fourier_matrix(graph, inv_F=None):
        inv_F = inv_F.toarray()
        # compute fourier matrix via moebius function ((5) in Causal Signal Processing Paper)
        nodes = list(graph.nodes)
        F = np.identity(graph.number_of_nodes())
        for x1 in range(graph.number_of_nodes()):
            print(x1)
            for x2 in range(x1):
                if inv_F[x1][x2] != 0:
                    F[x1][x2] = - Transformer.get_moebius_sum(F, inv_F, x1, x2)
        return F

    def get_fourier_coefficients(activations, inv_F=None, type='standard', lu_piv=None):
        signals = activations.to_vector()

        if type == 'standard':
            fourier_signals = scipy.sparse.linalg.spsolve_triangular(inv_F, signals)
        elif type == 'laplacian':
            fourier_signals = scipy.linalg.lu_solve(lu_piv, signals)
        else:
            raise NotImplementedError

        fourier_coefficients = activations.to_activations(fourier_signals)
        return fourier_coefficients



import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import os

class Visualizer:
    def get_min_max(matrix):
        vmin = min(matrix.view(-1))
        vmax = max(matrix.view(-1))
        return vmin, vmax

    def get_global_min_max(matrix_list):
        vmin = 0
        vmax = 0
        for m in matrix_list:
            new_vmin, new_vmax = Visualizer.get_min_max(m)
            vmin = min(vmin, new_vmin)
            vmax = max(vmax, new_vmax)
        return vmin, vmax

    def get_norm(matrix_list):
        vmin, vmax = Visualizer.get_global_min_max(matrix_list)
        absmax = max(abs(vmin), abs(vmax))
        if absmax == 0.:
            absmax = 1.e-307
        vmin = -absmax
        vmax = absmax
        norm = matplotlib.colors.TwoSlopeNorm(vmin=vmin, vcenter=0., vmax=vmax)
        return norm

    def visualize_pattern(activations, pdf_filepath, style='globalnorm', cmap_style='viridis'):
        os.makedirs(os.path.dirname(pdf_filepath), exist_ok=True)
        with PdfPages(pdf_filepath) as pdf:
            globalnorm = Visualizer.get_norm(activations.layeractivations)
            globalvmin, globalvmax = Visualizer.get_global_min_max(activations.layeractivations)

            for index_layer, (layeractivation, layername) in enumerate(zip(activations.layeractivations, activations.layernames)):
                layervmin, layervmax = Visualizer.get_global_min_max(layeractivation)
                layernorm = Visualizer.get_norm(layeractivation)

                max_cols = 4
                if (len(layeractivation.shape) == 1):
                    layeractivation  = layeractivation.unsqueeze(0).unsqueeze(0)
                n_channels = layeractivation.shape[0]
                n_cols = min(n_channels, max_cols)

                fig, axes = plt.subplots(-(-n_channels // max_cols), n_cols, sharey=True, subplot_kw={'xticks': []})
                fig.suptitle(f'Layer {index_layer}: {layername}')
                #fig.suptitle(f'Layer {index_layer+1}')
                for i, ax in enumerate(np.array(axes).reshape(-1)):
                    if i < n_channels:
                        if style == 'standard':
                            cmap = ax.imshow(layeractivation[i], aspect='equal', cmap=plt.get_cmap(cmap_style))
                        elif style == 'layernorm':
                            cmap = ax.imshow(layeractivation[i], aspect='equal', cmap=plt.get_cmap(cmap_style), norm=layernorm)
                        elif style == 'layerscale':
                            cmap = ax.imshow(layeractivation[i], aspect='equal', cmap=plt.get_cmap(cmap_style), vmin=layervmin, vmax=layervmax)
                        elif style == 'globalnorm':
                            cmap = ax.imshow(layeractivation[i], aspect='equal', cmap=plt.get_cmap(cmap_style), norm=globalnorm)
                        elif style == 'globalscale':
                            cmap = ax.imshow(layeractivation[i], aspect='equal', cmap=plt.get_cmap(cmap_style), vmin=globalvmin, vmax=globalvmax)
                        elif style == 'globalscalePositive':
                            cmap = ax.imshow(layeractivation[i], aspect='equal', cmap=plt.get_cmap(cmap_style), vmin=0, vmax=globalvmax)
                        else:
                            raise NotImplementedError
                    else:
                        ax.axis('off')
                cax = fig.add_axes([0.2,0.05,0.6,0.02])
                cbar = fig.colorbar(cmap, cax=cax, orientation='horizontal')
                cbar.ax.tick_params(labelsize=10)

                pdf.savefig(fig, bbox_inches='tight', pad_inches=0.5)
                plt.close()
